"""Define S3Uploader class."""
import os

from boto3 import client


class S3Uploader:
    """S3 uploader files class."""

    def __init__(self):
        """Init creds."""
        self.default_url = os.environ.get("MLFLOW_S3_ENDPOINT_URL", "http://minio-mlmanagement.ru/")
        self.default_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID", "minioadmin")
        self.default_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY", "minioadmin")

    def upload(self, local_path: str, bucket: str):
        """Upload."""
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileExistsError(f"Path: {local_path} does not exist")

        service_client = client(
            service_name="s3",
            use_ssl=True,
            verify=None,
            endpoint_url=self.default_url,
            aws_access_key_id=self.default_access_key_id,
            aws_secret_access_key=self.default_secret_access_key,
        )
        buckets = [item["Name"] for item in service_client.list_buckets()["Buckets"]]
        if bucket not in buckets:
            service_client.create_bucket(Bucket=bucket)

        if os.path.isfile(local_path):
            with open(local_path, "rb") as _file:
                service_client.upload_fileobj(_file, bucket)
        else:
            for root, _, files in os.walk(local_path):
                for file in files:
                    filename = file if root == local_path else os.path.join(os.path.relpath(root, start=local_path), file)
                    with open(os.path.join(root, file), "rb") as _file:
                        service_client.upload_fileobj(_file, bucket, filename)



"""Define train executor class."""
from abc import ABC

from ML_management.executor_template.executor_pattern import JobExecutorPattern
from ML_management.executor_template.upload_model_mode import UploadModelMode
from ML_management.models.model_type_to_methods_map import ModelMethodName


class TrainExecutor(JobExecutorPattern, ABC):
    """Train executor from pattern with defined settings parameters."""

    def __init__(self):
        super().__init__(
            executor_name="train",
            desired_model_methods=[ModelMethodName.train_function],
            executor_upload_model_mode=UploadModelMode.new_model,
        )

    def execute(self):
        """Define execute function that calls train_function of model with corresponding params."""
        return self.model.train_function(**self.model_methods_parameters[ModelMethodName.train_function])



"""Common variables."""
import os

server_ml_api = os.environ.get("server_url", "http://localhost:8000") + "/mlflow"
is_server = os.environ.get("is_server", False)

active_run_stack = []
active_experiment_name = None

EXPERIMENT_NAME_FOR_EXECUTOR = "executors"
EXPERIMENT_NAME_FOR_DATASET_LOADER = "dataset_loaders"



"""Dataset loader template for custom dataset loader."""
from abc import ABC, abstractmethod

from ML_management.mlmanagement import mlmanagement, utils
from ML_management.mlmanagement.utils import EXPERIMENT_NAME_FOR_DATASET_LOADER
from mlflow.pyfunc import PythonModel


class DatasetLoaderPattern(PythonModel, ABC):
    """Define dataset loader."""

    def __init__(self, dataset_loader_name: str):
        """
        Init dataset loader class.

        :param dataset_loader_name: The name of the dataset loader
        """
        self.dataset_loader_name = dataset_loader_name

        # That parameters will be set automatically in job before the 'get_dataset' func would be executed.
        self.data_path = None

    @abstractmethod
    def get_dataset(self, **dataset_params):
        """
        Return dataset.

        To get data_path use self.data_path parameter, which also will be set in the job.
        'dataset_params' are dataset_loader parameters. One has to define it as ordinary kwargs
        with type annotation.
        """
        raise NotImplementedError

    def upload_dataset_loader(self, pip_requirements=None, extra_pip_requirements=None, conda_env=None):
        """
        Upload wrapper to MLmanagement server.

        :param pip_requirements: {{ pip_requirements }}
        :param extra_pip_requirements: {{ extra_pip_requirements }}
        `pip_requirements` and 'extra_pip_requirements' must be either a string path to a pip requirements file on the
            local filesystem or an iterable of pip requirement strings.
        :param conda_env: {{ conda_env }}
        'conda_env' must be a dict specifying the conda environment for this model.
        """
        old_experiment_name = utils.active_experiment_name
        mlmanagement.set_experiment(EXPERIMENT_NAME_FOR_DATASET_LOADER)
        try:
            with mlmanagement.start_run(nested=True):
                mlmanagement.log_model(
                    artifact_path="",
                    python_model=self,
                    registered_model_name=self.dataset_loader_name,
                    pip_requirements=pip_requirements,
                    extra_pip_requirements=extra_pip_requirements,
                    conda_env=conda_env,
                )
        except Exception as err:
            raise err
        finally:
            utils.active_experiment_name = old_experiment_name



import os
import shutil
import unittest
from unittest.mock import Mock, patch

import boto3
import ML_management.collectors.s3.s3collector
from botocore.exceptions import ClientError
from ML_management.collectors import COLLECTORS
from ML_management.collectors.s3.s3collector import S3BucketNotFound, S3FolderNotFound, S3ObjectNotFound


def mock_client_download_successful(*args, **kwargs):
    client = Mock(spec=boto3.client)

    def mock_download_file(Bucket, Key, Fileobj, ExtraArgs=None, Callback=None, Config=None):
        files_content_map = {
            "sample_data/1.txt": "one\n",
            "sample_data/2.txt": "two\n",
            "sample_data/sample_folder/3.txt": "three\n",
        }
        Fileobj.write(files_content_map[Key].encode("utf-8"))

    client.download_fileobj = Mock(side_effect=mock_download_file)
    return client


def mock_client_download_folder_successful(*args, **kwargs):
    client = Mock(spec=boto3.client)

    class mock_paginator:
        def paginate(self, *args, **kwargs):
            return [
                {
                    "KeyCount": 3,
                    "Contents": [
                        {"Key": "sample_data/1.txt"},
                        {"Key": "sample_data/2.txt"},
                        {"Key": "sample_data/sample_folder/3.txt"},
                    ],
                }
            ]

    def mock_download_file(Bucket, Key, Fileobj, ExtraArgs=None, Callback=None, Config=None):
        files_content_map = {
            "sample_data/1.txt": "one\n",
            "sample_data/2.txt": "two\n",
            "sample_data/sample_folder/3.txt": "three\n",
        }
        Fileobj.write(files_content_map[Key].encode("utf-8"))

    client.download_fileobj = Mock(side_effect=mock_download_file)
    client.get_paginator = Mock(return_value=mock_paginator())
    return client


def mock_client_bad_folder(*args, **kwargs):
    client = Mock(spec=boto3.client)

    class mock_paginator:
        def paginate(self, *args, **kwargs):
            return [{"KeyCount": 0}]

    client.get_paginator = Mock(return_value=mock_paginator())
    return client


def mock_client_bad_object(*args, **kwargs):
    client = Mock(spec=boto3.client)

    class mock_paginator:
        def paginate(self, *args, **kwargs):
            return [{"KeyCount": 0}]

    client.download_fileobj = Mock(side_effect=ClientError({"Error": {"Code": "404"}}, "download_fileobj"))
    client.get_paginator = Mock(return_value=mock_paginator())
    client.head_bucket = Mock()
    return client


def mock_client_bad_bucket_folder(*args, **kwargs):
    client = Mock(spec=boto3.client)

    class mock_paginator:
        def paginate(self, *args, **kwargs):
            raise ClientError({"Error": {"Code": "NoSuchBucket"}}, "list_objects_v2")

    client.get_paginator = Mock(return_value=mock_paginator())

    return client


def mock_client_bad_bucket_object(*args, **kwargs):
    client = Mock(spec=boto3.client)

    class mock_paginator:
        def paginate(self, *args, **kwargs):
            return [{"KeyCount": 0}]

    client.download_fileobj = Mock(side_effect=ClientError({"Error": {"Code": "404"}}, "download_fileobj"))
    client.get_paginator = Mock(return_value=mock_paginator())
    client.head_bucket = Mock(side_effect=ClientError({"Error": {"Code": "404"}}, "download_fileobj"))
    return client


class TestS3Dataset(unittest.TestCase):
    """
    Mock S3 structure:

    test-data:
        sample_data/1.txt ("one\n"),
        sample_data/2.txt ("two\n"),
        sample_data/sample_folder/3.txt ("three\n")
    """

    @patch(
        "ML_management.collectors.s3.s3collector.client",
        new=mock_client_download_successful,
    )
    def test_download_files(self):
        s3_dataset = COLLECTORS["s3"]()
        assert mock_client_download_successful is ML_management.collectors.s3.s3collector.client
        endpoint_url = "some_mock_url"

        local_path = s3_dataset.set_data(
            local_path=os.path.join(os.path.dirname(__file__), "downloaded_data/"),
            bucket="test-data",
            remote_paths=[
                "sample_data/1.txt",
                "sample_data/2.txt",
                "sample_data/sample_folder/3.txt",
            ],
            endpoint_url=endpoint_url,
            aws_access_key_id="some_mock_key",
            aws_secret_access_key="some_mock_key",
        )

        with open(os.path.join(local_path, "sample_data/1.txt"), "r") as f:
            text_one = f.read()
            self.assertEqual(text_one, "one\n")
        with open(os.path.join(local_path, "sample_data/2.txt"), "r") as f:
            text_one = f.read()
            self.assertEqual(text_one, "two\n")
        with open(os.path.join(local_path, "sample_data/sample_folder/3.txt"), "r") as f:
            text_one = f.read()
            self.assertEqual(text_one, "three\n")

        shutil.rmtree(os.path.join(local_path, "sample_data/"))

    @patch(
        "ML_management.collectors.s3.s3collector.client",
        new=mock_client_download_folder_successful,
    )
    def test_download_folder(self):
        s3_dataset = COLLECTORS["s3"]()
        assert mock_client_download_folder_successful is ML_management.collectors.s3.s3collector.client
        endpoint_url = "some_mock_url"

        local_path = s3_dataset.set_data(
            local_path=os.path.join(os.path.dirname(__file__), "downloaded_data/"),
            bucket="test-data",
            remote_paths=["sample_data/"],
            endpoint_url=endpoint_url,
            aws_access_key_id="some_mock_key",
            aws_secret_access_key="some_mock_key",
        )

        with open(os.path.join(local_path, "sample_data/1.txt"), "r") as f:
            text_one = f.read()
            self.assertEqual(text_one, "one\n")
        with open(os.path.join(local_path, "sample_data/2.txt"), "r") as f:
            text_one = f.read()
            self.assertEqual(text_one, "two\n")
        with open(os.path.join(local_path, "sample_data/sample_folder/3.txt"), "r") as f:
            text_one = f.read()
            self.assertEqual(text_one, "three\n")

        shutil.rmtree(os.path.join(local_path, "sample_data/"))

    @patch(
        "ML_management.collectors.s3.s3collector.client",
        new=mock_client_bad_object,
    )
    def test_download_bad_object(self):
        s3_dataset = COLLECTORS["s3"]()
        assert mock_client_bad_object is ML_management.collectors.s3.s3collector.client
        endpoint_url = "some_mock_url"

        with self.assertRaises(S3ObjectNotFound):
            s3_dataset.set_data(
                local_path=os.path.join(os.path.dirname(__file__), "downloaded_data/"),
                bucket="test-data",
                remote_paths=["sample_data/nosuchfile.txt"],
                endpoint_url=endpoint_url,
                aws_access_key_id="some_mock_key",
                aws_secret_access_key="some_mock_key",
            )

    @patch(
        "ML_management.collectors.s3.s3collector.client",
        new=mock_client_bad_folder,
    )
    def test_download_bad_folder(self):
        s3_dataset = COLLECTORS["s3"]()
        assert mock_client_bad_folder is ML_management.collectors.s3.s3collector.client
        endpoint_url = "some_mock_url"

        with self.assertRaises(S3FolderNotFound):
            s3_dataset.set_data(
                local_path=os.path.join(os.path.dirname(__file__), "downloaded_data/"),
                bucket="test-data",
                remote_paths=["sample_data/nosuchfolder/"],
                endpoint_url=endpoint_url,
                aws_access_key_id="some_mock_key",
                aws_secret_access_key="some_mock_key",
            )

    @patch(
        "ML_management.collectors.s3.s3collector.client",
        new=mock_client_bad_bucket_folder,
    )
    def test_download_bad_bucket_folder(self):
        s3_dataset = COLLECTORS["s3"]()
        assert mock_client_bad_bucket_folder is ML_management.collectors.s3.s3collector.client
        endpoint_url = "some_mock_url"

        with self.assertRaises(S3BucketNotFound):
            s3_dataset.set_data(
                local_path=os.path.join(os.path.dirname(__file__), "downloaded_data/"),
                bucket="nosuchbucket",
                remote_paths=["sample_data/"],
                endpoint_url=endpoint_url,
                aws_access_key_id="some_mock_key",
                aws_secret_access_key="some_mock_key",
            )

    @patch(
        "ML_management.collectors.s3.s3collector.client",
        new=mock_client_bad_bucket_object,
    )
    def test_download_bad_bucket_object(self):
        s3_dataset = COLLECTORS["s3"]()
        assert mock_client_bad_bucket_object is ML_management.collectors.s3.s3collector.client
        endpoint_url = "some_mock_url"

        with self.assertRaises(S3BucketNotFound):
            s3_dataset.set_data(
                local_path=os.path.join(os.path.dirname(__file__), "downloaded_data/"),
                bucket="nosuchbucket",
                remote_paths=["sample_data/1.txt"],
                endpoint_url=endpoint_url,
                aws_access_key_id="some_mock_key",
                aws_secret_access_key="some_mock_key",
            )


if __name__ == "__main__":
    unittest.main()



"""S3 Collector for downloading files and folders."""
import os
import posixpath
from typing import List, Optional, Union

from boto3 import client
from botocore.exceptions import ClientError
from ML_management.collectors.collector_pattern import CollectorPattern


class S3FolderNotFound(Exception):
    """Define Version Not Found Exception."""

    def __init__(self, path: str, bucket: str):
        self.path = path
        self.bucket = bucket
        self.message = f'Folder "{path}" is not found in "{bucket}" bucket'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3FolderNotFound, (self.path, self.bucket))


class S3BucketNotFound(Exception):
    """Define Bucket Not Found Exception."""

    def __init__(self, bucket: str):
        self.bucket = bucket
        self.message = f'Bucket "{bucket}" does not exist'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3FolderNotFound, (self.bucket))


class S3ObjectNotFound(Exception):
    """Define Version Not Found Exception."""

    def __init__(self, path: str, bucket: str):
        self.path = path
        self.bucket = bucket
        self.message = f'Object "{path}" is not found in "{bucket}" bucket'
        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (S3ObjectNotFound, (self.path, self.bucket))


class S3Collector(CollectorPattern):
    """Collector for S3 paths using boto3 library."""

    def __init__(self) -> None:
        self.default_url = os.environ.get("MLFLOW_S3_ENDPOINT_URL", "http://localhost:9000")
        self.default_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID", "minioadmin")
        self.default_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY", "minioadmin")

    @staticmethod
    def get_json_schema():
        """Return json schema."""
        return {
            "type": "object",
            "properties": {
                "local_path": {"type": "string"},
                "bucket": {"type": "string"},
                "remote_paths": {"type": "array", "items": {"type": "string"}},
                "service_name": {"type": "string"},
                "region_name": {"type": "string"},
                "api_version": {"type": "string"},
                "use_ssl": {"type": "boolean"},
                "verify": {
                    "anyOf": [
                        {"type": "boolean"},
                        {"type": "string"},
                    ]
                },
                "endpoint_url": {"type": "string"},
                "aws_access_key_id": {"type": "string"},
                "aws_secret_access_key": {"type": "string"},
                "aws_session_token": {"type": "string"},
            },
            "required": ["bucket"],
            "additionalProperties": False,
        }

    def _download_file(self, bucket: str, service_client, remote_file_path: str, local_path: str):
        dirpath = posixpath.dirname(remote_file_path)
        local_dir_path = os.path.join(local_path, dirpath)
        local_file_path = os.path.join(local_path, remote_file_path)
        if not os.path.exists(local_dir_path):
            os.makedirs(local_dir_path, exist_ok=True)
        try:
            with open(local_file_path, "wb") as _file:
                service_client.download_fileobj(bucket, remote_file_path, _file)
        except ClientError as err:
            if err.response["Error"]["Code"] == "404":
                # maybe user forgot to add trailing slash? Try to download as folder
                try:
                    remote_folder_path = remote_file_path + "/"
                    self._download_folder(bucket, service_client, remote_folder_path, local_path)
                except S3FolderNotFound:
                    try:
                        # maybe, bad bucket?
                        service_client.head_bucket(Bucket=bucket)
                    except ClientError as err:
                        if err.response["Error"]["Code"] == "404":
                            # bad bucket it is.
                            raise S3BucketNotFound(bucket=bucket)
                        else:
                            raise err
                    # bucket is ok, it is a bad object
                    raise S3ObjectNotFound(path=remote_file_path, bucket=bucket)  # attempt of search as a folder failed
            else:
                raise err
        return local_file_path

    def _download_folder(self, bucket, service_client, folder_path, local_path):
        paginator = service_client.get_paginator("list_objects_v2")
        # TODO prevent infinite loop, sometimes the dir is recursively included ???
        try:
            page_iterator = paginator.paginate(Bucket=bucket, Prefix=folder_path)
            for page in page_iterator:
                if page["KeyCount"] == 0:
                    # folder not found or is empty?
                    raise S3FolderNotFound(path=folder_path, bucket=bucket)
                # Objects listed directly will be files
                for obj in page.get("Contents", []):
                    file_path = obj.get("Key")
                    self._download_file(bucket, service_client, file_path, local_path)
        except ClientError as err:
            if err.response["Error"]["Code"] == "NoSuchBucket":
                raise S3BucketNotFound(bucket=bucket)
            else:
                raise err

    def set_data(
        self,
        *,
        local_path: str = "/s3_data/",
        bucket: str,  # TODO do we pass bucket or not? is it always possible to parse?
        remote_paths: Optional[List[str]] = None,
        service_name: str = "s3",
        region_name: Optional[str] = None,
        api_version: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[Union[bool, str]] = None,
        endpoint_url: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
    ) -> str:
        """
        Set data.

        :type local_path: string
        :param local_path: Local path to save data to.  Defaults to /s3_data/.

        :type bucket: string
        :param bucket: Bucket containing requested files.

        :type remote_paths: list(string)
        :param remote_paths: List of paths relative to passed bucket.  Each path
            can represent either a single file, or a folder.  If a path represents
            a folder (should end with a slash), then all contents of a folder are recursively downloaded.

        :type service_name: string
        :param service_name: The name of a service, e.g. 's3' or 'ec2'
        available to boto3.  Defaults to 's3'.

        :type region_name: string
        :param region_name: The name of the region associated with the client.
            A client is associated with a single region.

        :type api_version: string
        :param api_version: The API version to use.  By default, botocore will
            use the latest API version when creating a client.  You only need
            to specify this parameter if you want to use a previous API version
            of the client.

        :type use_ssl: boolean
        :param use_ssl: Whether to use SSL.  By default, SSL is used.
            Note that not all services support non-ssl connections.

        :type verify: boolean/string
        :param verify: Whether to verify SSL certificates.  By default,
            SSL certificates are verified.  You can provide the following
            values:

            * False - do not validate SSL certificates.  SSL will still be
              used (unless use_ssl is False), but SSL certificates
              will not be verified.
            * path/to/cert/bundle.pem - A filename of the CA cert bundle to
              uses.  You can specify this argument if you want to use a
              different CA cert bundle than the one used by botocore.

        :type endpoint_url: string
        :param endpoint_url: The complete URL to use for the constructed
            client. Normally, botocore will automatically construct the
            appropriate URL to use when communicating with a service.  You
            can specify a complete URL (including the "http/https" scheme)
            to override this behavior.  If this value is provided,
            then ``use_ssl`` is ignored.

        :type aws_access_key_id: string
        :param aws_access_key_id: The access key to use when creating
            the client.  This is entirely optional, and if not provided,
            the credentials configured for the session will automatically
            be used.  You only need to provide this argument if you want
            to override the credentials used for this specific client.

        :type aws_secret_access_key: string
        :param aws_secret_access_key: The secret key to use when creating
            the client.  Same semantics as aws_access_key_id above.

        :type aws_session_token: string
        :param aws_session_token: The session token to use when creating
            the client.  Same semantics as aws_access_key_id above.
        """
        service_client = client(
            service_name=service_name,
            region_name=region_name,
            api_version=api_version,
            use_ssl=use_ssl,
            verify=verify,
            endpoint_url=endpoint_url if endpoint_url else self.default_url,
            aws_access_key_id=aws_access_key_id if aws_access_key_id else self.default_access_key_id,
            aws_secret_access_key=aws_secret_access_key if aws_secret_access_key else self.default_secret_access_key,
            aws_session_token=aws_session_token,
        )
        if remote_paths is not None:
            for path in remote_paths:
                if path.endswith("/"):
                    # it is a folder
                    # list all files in a folder in one go and download them all!
                    self._download_folder(bucket, service_client, path, local_path)
                else:
                    # it is a file
                    self._download_file(bucket, service_client, path, local_path)
        else:
            for key in service_client.list_objects(Bucket=bucket)["Contents"]:
                self._download_file(bucket, service_client, key["Key"], local_path)
        return local_path



"""Abstract class of Collector for Job."""
from abc import ABC, abstractmethod

# That's copied from server, excluding splitters


class CollectorPattern(ABC):
    """Abstract class of Collector for Job."""

    @abstractmethod
    def set_data(self, *args, **kwargs):
        """Set suitable data."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_json_schema():
        """
        Every dataset should define json schema for its set_data parameters.

        https://json-schema.org/draft-07/json-schema-validation.html#rfc.section.6.1.1

        The value of keyword "type" MUST be either a string or an array.
        If it is an array, elements of the array MUST be strings and MUST be unique.

        String values MUST be one of the six primitive types ("null", "boolean", "object", "array", "number", or "string"),
        or "integer" which matches any number with a zero fractional part.

        Parameters are optional by default. Requiered parameters are specified as array by key "required".
        """
        raise NotImplementedError



"""Custom exception definition for jsonschema inference."""
from typing import List


class DictKeysMustBeStrings(Exception):
    """Define DictKeysMustBeStrings exception."""

    def __init__(self, annotation: str):
        self.annotation = annotation
        self.message = "Dict keys must be strings: {annotation}".format(annotation=annotation)

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (DictKeysMustBeStrings, (self.annotation))


class InvalidSchema(Exception):
    """Define InvalidSchema exception."""

    def __init__(self, schema: dict, original_message: str):
        self.schema = schema
        self.original_message = original_message
        self.message = "Inferred jsonschema is not valid. Please, report this bug \
            to developers. Schema: {schema}, original exception message: {original_message}".format(
            schema=schema, original_message=original_message
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidSchema, (self.schema, self.original_message))


class NoAnnotation(Exception):
    """Define NoAnnotation exception."""

    def __init__(self, arg: str):
        self.arg = arg
        self.message = "Argument '{arg}' has no type annotation".format(arg=arg)

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (NoAnnotation, (self.arg))


class UnsupportedType(Exception):
    """Define UnsupportedType exception."""

    def __init__(self, annotation: str, supported_types: List[str]):
        self.annotation = annotation
        self.supported_types = supported_types
        self.message = "Only {supported_types} and typing.List, typing.Dict, \
            typing.Union and typing.Optional are supported. Cause: {annotation}".format(
            annotation=annotation, supported_types=", ".join(supported_types)
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (UnsupportedType, (self.annotation, self.supported_types))


class FunctionContainsVarArgs(Exception):
    """Define FunctionContainsVarArgs exception."""

    def __init__(self, function_name: str):
        self.function_name = function_name
        self.message = "Function '{function_name}' contains a variable number positional arguments i.e. *args".format(
            function_name=function_name
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (FunctionContainsVarArgs, (self.function_name))


class FunctionContainsVarKwArgs(Exception):
    """Define FunctionContainsVarKwArgs exception."""

    def __init__(self, function_name: str):
        self.function_name = function_name
        self.message = "Function '{function_name}' contains a variable number keyword arguments i.e. **kwargs".format(
            function_name=function_name
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (FunctionContainsVarKwArgs, (self.function_name))



"""Define test executor class."""
from abc import ABC

from ML_management.executor_template.executor_pattern import JobExecutorPattern
from ML_management.executor_template.upload_model_mode import UploadModelMode
from ML_management.models.model_type_to_methods_map import ModelMethodName


class TestExecutor(JobExecutorPattern, ABC):
    """Test executor from pattern with defined settings parameters."""

    def __init__(self):
        super().__init__(
            executor_name="test", desired_model_methods=[ModelMethodName.predict_function], executor_upload_model_mode=UploadModelMode.none
        )

    def execute(self):
        """Define execute function that calls predict_function of model with corresponding params."""
        return self.model.predict_function(**self.model_methods_parameters[ModelMethodName.predict_function])



"""Class inherits from base class of Model with a specific method for torch module."""
from abc import ABC, abstractmethod

from ML_management.models.patterns import model_pattern


class TorchModel(model_pattern.Model, ABC):
    """Implementation of torch model."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_nn_module(self, **kwargs):
        """Define nn_module."""
        raise NotImplementedError



"""Setup for ml-management package."""
from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("VERSION", "r", encoding="utf-8") as f:
    version = f.read()

setup(
    name="ml-management",
    version=version,
    description="Python implementation of model pattern, dataset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="ISPRAS MODIS",
    author_email="modis@ispras.ru",
    maintainer="Maxim Ryndin",
    packages=find_packages(include=["ML_management", "ML_management.*"]),
    include_package_data=True,
    # jsonschema 2.6.0 is the last one to support python 3.6
    install_requires=[
        "sgqlc==16.1",
        "mlflow==1.21.0",
        "pandas>=0.20.1,<=1.5.2",
        "numpy>=1.8.1,<=1.23.5",
        "jsonschema==2.6.0",
        "requests_toolbelt>=0.9.1,<=0.10.1",
        "boto3==1.21.21",
        "protobuf<4.0.0",
    ],
    data_files=[("", ["VERSION"])],
    python_requires=">=3.6",
)



"""Define Model Registry Manager."""
from functools import partial
from operator import attrgetter

from ML_management import mlmanagement
from ML_management.registry.exceptions import MetricNotLogged, ModelNotRegistered, NoMetricProvided, UnsupportedCriteria, VersionNotFound
from mlflow.entities.model_registry.model_version import ModelVersion
from mlflow.entities.model_registry.registered_model import RegisteredModel
from mlflow.exceptions import RestException
from mlflow.store.entities import PagedList

# noinspection PyUnresolvedReferences
# TODO maybe make it singleton?


class RegistryManager:
    """Registry Manager to choose necessary version of the model."""

    def __init__(self):
        self.client = mlmanagement.MlflowClient()

    def check_model_registered(self, name: str) -> RegisteredModel:
        """Check if model is registered in mlflow or not."""
        try:
            registered_model = self.client.get_registered_model(name)
        except RestException as err:
            if err.error_code == "RESOURCE_DOES_NOT_EXIST":
                raise ModelNotRegistered(name)
            else:
                raise err
        return registered_model

    def get_latest_version(self, name: str) -> ModelVersion:
        """Get the latest version of a model "name"."""
        latest_version = None
        registered_model = self.check_model_registered(name)

        # if there is only one default stage, there should be only one latest version
        # otherwise something is wrong on backend
        # version is not supposed to have no versions
        latest_version = registered_model.latest_versions[0]

        return latest_version

    def get_best_version(
        self,
        name: str,
        metric: str,
        optimal_min: bool = False,
    ) -> ModelVersion:
        """Get best version of model "name" according to a metric."""
        # optimal_min parameter is to look for minimal value of metric, max by default
        current_best_version, current_best_score = None, None
        self.check_model_registered(name)
        # MlflowClient.search_model_versions always returns all versions in single page with no token
        # as it uses SqlAlchemyStore on mlflow server pod based on helm Values.mlflow.backendUri
        # which starts with "postgres". See mlflow/store/model_registry/sqlalchemy_store.py:732
        model_versions = self.client.search_model_versions(filter_string=f"name = '{name}'")

        for mv in model_versions:
            try:
                metric_value = self.client.get_run(run_id=mv.run_id).data.metrics[metric]
            except KeyError:
                continue  # metric might not be logged for SOME versions, so don't raise straight away

            if current_best_score is None:
                current_best_version, current_best_score = mv, metric_value
            else:
                if optimal_min:
                    if metric_value < current_best_score:
                        current_best_version, current_best_score = mv, metric_value
                else:
                    if metric_value > current_best_score:
                        current_best_version, current_best_score = mv, metric_value
        if current_best_version:
            return current_best_version
        else:
            raise MetricNotLogged(name, metric)

    def get_initial_version(self, name: str) -> ModelVersion:
        """Get initial version of model "name"."""
        # TODO make faster initial version retrieval through Model table at server
        model_versions = self.get_all_versions(name)
        min_version = min(model_versions, key=attrgetter("version"))

        return min_version

    def get_version(self, name: str, version: int) -> ModelVersion:
        """
        Get model version from MLflow Model Registry and return its ModelVersion object.

        https://mlflow.org/docs/latest/python_api/mlflow.entities.html#mlflow.entities.model_registry.ModelVersion

        Parameters:
            name (str): Model name used for model registration.
            version (int): Desired version number.

        Returns:
            ModelVersion: Desired model version

        """
        # first, check that model is registered
        self.check_model_registered(name)
        # now, try to retrieve desired version
        try:
            model_version = self.client.get_model_version(name, str(version))
        except RestException as err:
            if err.error_code == "RESOURCE_DOES_NOT_EXIST":
                # model is registered, but version is not found
                raise VersionNotFound(name, str(version))
            else:
                raise err

        return model_version

    def get_all_versions(
        self,
        name: str,
    ) -> PagedList[ModelVersion]:
        """
        Get all versions of a given model and return PagedList of ModelVersion objects.

        https://mlflow.org/docs/latest/python_api/mlflow.entities.html#mlflow.entities.model_registry.ModelVersion

        Parameters:
            name (str): Model name used for model registration.

        Returns:
            PagedList[ModelVersion]: Available model versions
        """
        self.check_model_registered(name)

        # MlflowClient.search_model_versions always returns all versions in single page with no token
        # as it uses SqlAlchemyStore on mlflow server pod based on helm Values.mlflow.backendUri
        # which starts with "postgres". See mlflow/store/model_registry/sqlalchemy_store.py:732
        model_versions = self.client.search_model_versions(filter_string=f"name = '{name}'")

        return model_versions

    def choose_version(
        self,
        *,
        name: str,
        criteria: str,  # TODO enum?
        metric: str = None,
        optimal_min: bool = False,
    ) -> ModelVersion:
        """
        Choose optimal model version from MLflow Model Registry and return its ModelVersion object.

        https://mlflow.org/docs/latest/python_api/mlflow.entities.html#mlflow.entities.model_registry.ModelVersion

        Choice is made according to specified criteria.

        Parameters:
            name (str): Model name used for model registration.
            criteria (str): Criteria to choose between model versions. Must be one of: "initial", "latest", "best".
            metric (str): Metric to use with "best" criteria. Has no effect otherwise.
            optimal_min (bool): If set to True and "best" criteria is used, then choose version with minimal
                value of "metric" (useful if metric is a loss function). Defaults to False (choose version with maximal
                value of "metric").

        Returns:
            ModelVersion: Optimal model version

        """
        # for now, return single path and no name choosing -- model name must be explicitly passed
        # TODO in the future make name resolvers and return list of paths, one per chosen name
        version_fetcher_map = {
            "initial": self.get_initial_version,
            "latest": self.get_latest_version,
            "best": partial(
                self.get_best_version,
                metric=metric,
                optimal_min=optimal_min,
            ),
        }
        if criteria == "best" and metric is None:
            raise NoMetricProvided(criteria)
        if criteria in version_fetcher_map:
            return version_fetcher_map[criteria](name=name)
        else:
            raise UnsupportedCriteria(criteria, list(version_fetcher_map.keys))
